#!/usr/bin/env python
# coding: utf-8

# # custom_functions
# This notebook collects all the functions needed for the computation of the other notebooks.
# 
# Author: Luca Urbinati, 02-10-2019
# 
# Sources:
# - Chapter 5: Examples, pag 643, scikit-learn user guide, Release 0.20.3, Mar 01 2019

# ## Import libraries

# In[ ]:


from libraries import *

scale = 60 # 60 to get minutes, 1 to get seconds
time_scale = 'minutes'


# ## Functions

# In[ ]:


def adjusted_prediction(y_score, t): # TESTATO
    """
    This function adjusts the prediction based on the decision threshold t.
    It works with binary classifiers only.
    """

    return np.asarray([1 if y >= t else 0 for y in y_score])


# In[ ]:


def minimize_wasted(y_true, y_score): # TESTATO
    """
    Function that minimizes the non main diagonal of the confusion matrix and returns the best decision threshold t.
    """
    
    _, _, thresholds = precision_recall_curve(y_true, y_score)

    min_t = np.min(thresholds)
    max_t = np.max(thresholds)
    step_t = (max_t - min_t)/100 # 100 points in the thresholds interval
    print("min_t: %.3f, max_t: %.3f, step: %0.3f" % (min_t, max_t, step_t))
    
    minimum_wasted = 100
    best_t = 0
    
    for t in np.arange(min_t, max_t, step_t):
        y_pred_adj = adjusted_prediction(y_score, t)
        cm = confusion_matrix(y_true, y_pred_adj)
        wasted = (cm[0][1]+cm[1][0])/(cm[0][0]+cm[0][1]+cm[1][0]+cm[1][1]) * 100
        if(minimum_wasted > wasted):
            minimum_wasted = wasted       
            best_t = t
            
    print("minimum_wasted: %0.3f%%, best_t: %0.3f" % (minimum_wasted, best_t))
    
    return minimum_wasted, best_t


# In[ ]:


def grid_search_wrapper(clf, X, y, tuned_parameters, inner_cv, scorer, scorer_name, n_jobs=1):
    """
    This function fits a GridSearchCV classifier using refit_score when single scorer is used.
    X, y = development vectors
    """

    t0 = time.time()
    
    print("Grid creation")
    grid_search_clf = GridSearchCV(clf, param_grid=tuned_parameters, cv=inner_cv, scoring=scorer, iid=True, return_train_score=True, error_score=0, n_jobs=n_jobs)
    print("Fitting")
    grid_search_clf.fit(X, y)

    print()
    print("Best parameters: %s" % grid_search_clf.best_params_)
    print("Best non-nested score: %s" % grid_search_clf.best_score_)
    
    print("-"*100)
    elapsed = (time.time() - t0)/scale
    print("Elapsed time: %d minutes\n" % elapsed)
    
    return grid_search_clf


# In[ ]:


def confidence_interval(mean, std_dev, n_samples, quantile=None):
    """
    This function calculates the confidence interval at 95% level of confidence,
    given the mean (mean) and the standard deviation (std_dev) of the samples 
    and the number of the samples (n_samples).
    Since the variance of the population is not known, the Student's t-distribution is used.
    """
    quantiles = [0, 12.71,  4.303, 3.182, 2.776, 2.571, 2.447, 2.365, 2.306, 2.262]
    if quantile == None:
         quantile = quantiles[n_samples-1]
    std_error = std_dev/np.sqrt(n_samples)
    half_range = quantile*std_error
    
    return half_range


# In[ ]:


print("custom_functions imported")

